# frozen_string_literal: true

require "is/inspectable"
require "is/stateful"
require "securerandom"

module Core
  module Pipeline
    # [public] A pipeline action.
    #
    class Action
      include Is::Stateful
      state :__used_random_suffixes, default: []

      class << self
        # [public] Builds an action for a given target.
        #
        def build(target = nil, *args, before: nil, after: nil, context: nil, &block)
          if block
            Actions::Block.new(target, before: before, after: after, context: context, &block)
          else
            build_target(target, *args, before: before, after: after, context: context)
          end
        end

        private def build_target(first, *args, before:, after:, context:)
          case first
          when String, Symbol
            if (target = build_target(args[0], *args[1..], before: before, after: after, context: context))
              target
            else
              Actions::Method.new(first, before: before, after: after, context: context)
            end
          when NilClass
            nil
          else
            raise ArgumentError, "unsupported action type"
          end
        end

        def build_name
          suffix = generate_random_suffix
          if __used_random_suffixes.include?(suffix)
            build_name
          else
            __used_random_suffixes << suffix
            "action_#{suffix}"
          end
        end

        def generate_random_suffix
          SecureRandom.hex(8)
        end
      end

      include Is::Inspectable

      def initialize(name, before: nil, after: nil, context: nil)
        @name = (name || self.class.build_name).to_sym
        @before = before
        @after = after
        @context = context
      end

      # [public] The action name.
      #
      attr_reader :name

      # [public] The name of the action this action should be called before.
      #
      attr_reader :before

      # [public] The name of the action this action should be called after.
      #
      attr_reader :after

      # [public] Finalizes the action for the given context.
      #
      def finalize(context)
        raise "not implemented"
      end

      require_relative "actions/block"
      require_relative "actions/method"
    end
  end
end
