# frozen_string_literal: true

require "is/inspectable"

require_relative "action"
require_relative "compiler"

module Core
  module Pipeline
    include Is::Inspectable
    inspects without: [:@object, :@mutex]

    # [public] Calls a pipeline on a pipelined object.
    #
    class Callable
      def initialize(object)
        @object = object
        @mutex = Mutex.new
        @actions = []
        @skipped = []
      end

      attr_reader :actions
      attr_reader :skipped

      # [public] Returns true if the pipeline will call any actions.
      #
      def any?
        (@actions.map(&:name) - @skipped).any?
      end

      # [public] Defines a pipeline action.
      #
      def action(*args, before: nil, after: nil, context: nil, &block)
        @actions << Action.build(*args, before: before, after: after, context: context, &block)
      end

      # [public] Skips the given action.
      #
      def skip(name)
        @skipped << name.to_sym
      end

      # [public] Replaces existing actions with actions from the given pipeline.
      #
      def use(pipeline)
        if (pipeline.is_a?(Class) || pipeline.is_a?(Module)) && pipeline.ancestors.include?(Is::Pipeline)
          @mutex.synchronize do
            @actions.clear
            @actions.concat(pipeline.pipeline.actions)
          end
        else
          raise ArgumentError, "expected a pipeline"
        end
      end

      # [public] Includes actions from the given pipeline.
      #
      def include(pipeline)
        if (pipeline.is_a?(Class) || pipeline.is_a?(Module)) && pipeline.ancestors.include?(Is::Pipeline)
          @actions.concat(pipeline.pipeline.actions)
        else
          raise ArgumentError, "expected a pipeline"
        end
      end

      # [public] Excludes actions from the given pipeline.
      #
      def exclude(pipeline)
        if (pipeline.is_a?(Class) || pipeline.is_a?(Module)) && pipeline.ancestors.include?(Is::Pipeline)
          pipeline.pipeline.actions.each do |action|
            @actions.delete(action)
          end
        else
          raise ArgumentError, "expected a pipeline"
        end
      end

      # [public] Calls the pipeline in context of an object with the given arguments.
      #
      def call(object, ...)
        compile.call(object, ...)
      end

      private def compile
        instance_eval(Compiler.compile(self, @object), __FILE__, __LINE__ - 1)

        self
      end
    end
  end
end
