# frozen_string_literal: true

require "is/extension"
require "is/stateful"

require_relative "../core/pipeline/callable"
require_relative "../core/pipeline/compiler"
require_relative "../core/pipeline/signals/halted"
require_relative "../core/pipeline/signals/rejected"

module Is
  # [public] Turns any object into a pipeline with actions.
  #
  module Pipeline
    extend Is::Extension

    applies do |extended:|
      # [public] The current callable pipeline.
      #
      state :pipeline, default: Core::Pipeline::Callable.new(extended ? singleton_class : self)
    end

    extends dependencies: [
      Is::Stateful
    ]

    extends :definition do
      # [public] Defines a pipeline action.
      #
      def action(*args, before: nil, after: nil, context: nil, &block)
        pipeline.action(*args, before: before, after: after, context: context, &block)
      end

      # [public] Skips the given action.
      #
      def skip_action(name)
        pipeline.skip(name)
      end

      # [public] Replaces existing actions with actions from the given pipeline.
      #
      def use_pipeline(other_pipeline)
        pipeline.use(other_pipeline)
      end

      # [public] Includes actions from the given pipeline.
      #
      def include_pipeline(other_pipeline)
        pipeline.include(other_pipeline)
      end

      # [public] Excludes actions from the given pipeline.
      #
      def exclude_pipeline(other_pipeline)
        pipeline.exclude(other_pipeline)
      end
    end

    extends :implementation do
      # [public] Calls a pipeline with arguments.
      #
      def call(...)
        @pipeline.call(self, ...)
      end
    end

    extends :implementation, prepend: true do
      # [public] Halts the execution of the pipeline, setting the pipeline's current value to the given object.
      #
      private def halt(value = nil)
        raise Core::Pipeline::Signals::Halted.new(value)
      end

      # [public] Rejects the current action, skipping the rest of the action and calling remaining pipeline actions.
      #
      private def reject
        raise Core::Pipeline::Signals::Rejected
      end
    end
  end
end
